from typing import Any
import importlib
import time
import multiprocessing
import os
import json
from datetime import datetime, timezone
import logging

log = logging.getLogger(__name__)


def driver_api(
    driver: str, command: str, address: str, channel: int, **kwargs: dict
) -> Any:
    m = importlib.import_module(f"tomato.drivers.{driver}")
    func = getattr(m, command)
    return func(address, channel, **kwargs)


def data_poller(
    driver: str, address: str, channel: int, device: str, root: str, kwargs: dict
) -> None:
    pollrate = kwargs.pop("pollrate", 10)
    verbose = bool(kwargs.pop("verbose", 0))
    cont = True
    while cont:
        ts, nrows, data = driver_api(driver, "get_data", address, channel, **kwargs)
        while nrows > 0:
            isots = datetime.fromtimestamp(ts, tz=timezone.utc).isoformat()
            isots = isots.replace(":", "")
            fn = os.path.join(root, f"{device}_{isots}_data.json")
            with open(fn, "w") as of:
                json.dump(data, of)
            ts, nrows, data = driver_api(driver, "get_data", address, channel, **kwargs)

        ts, done, metadata = driver_api(
            driver, "get_status", address, channel, **kwargs
        )
        if verbose:
            isots = datetime.fromtimestamp(ts, tz=timezone.utc).isoformat()
            isots = isots.replace(":", "")
            fn = os.path.join(root, f"{device}_{isots}_status.json")
            with open(fn, "w") as of:
                json.dump(metadata, of)

        if done:
            cont = False
        else:
            time.sleep(pollrate)


def driver_worker(settings: dict, pipeline: dict, payload: dict, jobid: int) -> None:
    root = os.path.join(settings["queue"]["storage"], str(jobid))
    jobs = []
    for v in pipeline["devices"]:
        print(v)
        drv, addr, ch, tag = v["driver"], v["address"], v["channel"], v["tag"]
        dpar = settings["drivers"].get(drv, {})
        pl = payload["method"][tag]
        smpl = payload["sample"]

        log.debug(f"jobid {jobid}: getting status of device '{tag}'")
        ts, ready, metadata = driver_api(drv, "get_status", addr, ch, **dpar)
        assert ready, f"Failed: device '{tag}' is not ready."

        log.debug(f"jobid {jobid}: starting payload for device '{tag}'")
        start_ts = driver_api(drv, "start_job", addr, ch, **dpar, payload=pl, **smpl)
        metadata["uts"] = start_ts

        isots = datetime.fromtimestamp(ts, tz=timezone.utc).isoformat().replace(":", "")
        fn = os.path.join(root, f"{tag}_{isots}_status.json")
        with open(fn, "w") as of:
            json.dump(metadata, of)

        log.debug(f"jobid {jobid}: starting data polling for device '{tag}'")
        kwargs = dpar
        kwargs.update(
            {
                "pollrate": v.get("pollrate", 10),
                "verbose": v.get("verbose", 0),
            }
        )
        p = multiprocessing.Process(
            name=f"data_poller_{jobid}_{tag}",
            target=data_poller,
            args=(drv, addr, ch, tag, root, kwargs),
        )
        jobs.append(p)
        p.start()

    for p in jobs:
        p.join()
    return



from typing import Callable
import psutil
import os
import subprocess
import time
import json
import logging
import sys

log = logging.getLogger(__name__)

from ..drivers import driver_worker
from .. import dbhandler


def _find_matching_pipelines(pipelines: list, method: dict) -> list[str]:
    req_names = set(method.keys())
    req_capabs = []
    for k in req_names:
        for s in method[k]:
            req_capabs.append(s["name"])
    req_capabs = set(req_capabs)

    candidates = []
    for cd in pipelines:
        dnames = set([dev["tag"] for dev in cd["devices"]])
        if req_names.intersection(dnames) == req_names:
            candidates.append(cd)

    matched = []
    for cd in candidates:
        capabs = []
        for v in cd["devices"]:
            capabs += v["capabilities"]
        if req_capabs.intersection(set(capabs)) == req_capabs:
            matched.append(cd)

    return matched


def _pipeline_ready_sample(ret: tuple, sample: dict) -> bool:
    sampleid, ready, jobid, pid = ret
    if ready == 0:
        return False
    else:
        if sample["name"] == sampleid:
            return True
        else:
            return False


def job_wrapper() -> None:
    jobfile = sys.argv[1]
    with open(jobfile, "r") as infile:
        jsdata = json.load(infile)
    settings = jsdata["settings"]
    payload = jsdata["payload"]
    pipeline = jsdata["pipeline"]
    pip = pipeline["name"]
    jobid = jsdata["jobid"]
    queue = settings["queue"]
    state = settings["state"]
    pid = os.getpid()
    log.info(f"executing job '{jobid}' on pid '{pid}'")
    dbhandler.pipeline_assign_job(state["path"], pip, jobid, pid, type=state["type"])
    dbhandler.job_set_status(queue["path"], "r", jobid, type=queue["type"])
    dbhandler.job_set_time(queue["path"], "executed_at", jobid, type=queue["type"])
    driver_worker(settings, pipeline, payload, jobid)
    ready = payload.get("tomato", {}).get("unlock_when_done", False)
    dbhandler.job_set_status(queue["path"], "c", jobid, type=queue["type"])
    dbhandler.job_set_time(queue["path"], "completed_at", jobid, type=queue["type"])
    dbhandler.pipeline_reset_job(state["path"], pip, ready, type=state["type"])


def main_loop(settings: dict, pipelines: dict) -> None:
    qup = settings["queue"]["path"]
    qut = settings["queue"]["type"]
    stp = settings["state"]["path"]
    stt = settings["state"]["type"]
    while True:
        # check existing PIDs in state
        ret = dbhandler.pipeline_get_running(stp, type=stt)
        for pip, jobid, pid in ret:
            log.debug(f"checking PID of running job '{jobid}'")
            if psutil.pid_exists(pid) and "python" in psutil.Process(pid).name():
                log.debug(f"PID of running job '{jobid}' found")
                # dbhandler.job_set_status(queue, "r", jobid)
            else:
                log.debug(f"PID of running job '{jobid}' not found")
                dbhandler.pipeline_reset_job(stp, pip, False, type=stt)
                dbhandler.job_set_status(qup, "ce", jobid, type=qut)
                dbhandler.job_set_time(qup, "completed_at", jobid, type=qut)

        # check existing jobs in queue
        ret = dbhandler.job_get_all(qup, type=qut)
        for jobid, strpl, st in ret:
            payload = json.loads(strpl)
            if st in ["q", "qw"]:
                if st == "q":
                    log.debug(f"checking whether job '{jobid}' can ever be matched")
                matched_pips = _find_matching_pipelines(pipelines, payload["method"])
                if len(matched_pips) > 0 and st != "qw":
                    dbhandler.job_set_status(qup, "qw", jobid, type=qut)
                log.debug(f"checking whether job '{jobid}' can be queued")
                for pip in matched_pips:
                    pipinfo = dbhandler.pipeline_get_info(stp, pip["name"], type=stt)
                    can_queue = _pipeline_ready_sample(pipinfo, payload["sample"])
                    if can_queue:
                        dbhandler.pipeline_reset_job(stp, pip["name"], False, type=stt)
                        args = {
                            "settings": settings,
                            "pipeline": pip,
                            "payload": payload,
                            "jobid": jobid
                        }
                        root = os.path.join(settings["queue"]["storage"], str(jobid))
                        os.makedirs(root)
                        jpath = os.path.join(root, "jobdata.json")
                        with open(jpath, "w") as of:
                            json.dump(args, of, indent=1)
                        p = subprocess.Popen(
                            ["tomato_worker", str(jpath)], 
                            creationflags=subprocess.DETACHED_PROCESS | subprocess.CREATE_NO_WINDOW
                        )
                        break
        time.sleep(settings.get("main loop", 1))

        # - if jobid->status == q:
        #     find matching pipelines -> qw
        # - if jobid->status == qw:
        #     find matching pipelines
        #     find matching samples
        #     is pipeline ready -> r -> assign jobid and pid into pipeline state

    # for pname, pvals in pipelines.items():
    #    print(f'driver_worker(settings, pvals, None): with {pname}')
    #    driver_worker(settings, pvals, None)



""" Bio-Logic OEM package python API.

This module provides a pure Python interface to the EcLib DLL used to control Bio-Logic potentiostats.

As the methods of this API closely follow the DLL parameters, no docstring is provided,
as one can refer to the Development Package PDF for documentation.

The aim of this API is to shield this module's clients from the ctypes intricacies,
leaving the user to use either plain types or types coming from this module, or the kbio_types module.

The only consistent conventions in this API are :
  * id_ is the connection identifier returned by a Connect call,
  * ch is a 1 based channel identifier (vs a 0 based value in the DLL)
  * strings on the client side are encoded in this API, as the DLL uses bytes.

Most of the functions raise an exception on error (a BL_Error exception type),
which encapsulates the error code.

This behaviour can be overriden in the BL_xxx functions with an abort flag set to False.

"""

from array import array

from . import kbio_types as KBIO

from .c_utils import *
from .utils import pp_plural, warn_diff, exception_brief

# ==============================================================================#


class KBIO_api:
    def GetLibVersion(self):
        try:
            version = c_buffer(32)
            error = self.BL_GetLibVersion(*version.parm)
            return version.value
        except Exception as e:
            print(exception_brief(e, 1))

    def Connect(self, server, timeout=5):
        id_ = c_int32()
        info = self.DeviceInfo()
        self.BL_Connect(server.encode(), timeout, id_, info)
        # info is only provided by this call, so it must be kept by caller for further use.
        return id_.value, info

    def USB_DeviceInfo(self, index):

        company = c_buffer(128)
        device = c_buffer(128)
        serial_number = c_buffer(128)

        ok = self.BL_GetUSBdeviceinfos(
            index, *company.parm, *device.parm, *serial_number.parm
        )

        if not ok:
            raise RuntimeError(f"no information available for USB{index}")

        # fields must be cleaned of their NULL ending character
        return {
            "company": company.value[:-1],
            "device": device.value[:-1],
            "serial_number": serial_number.value[:-1],
        }

    def TestConnection(self, id_):
        error = self.BL_TestConnection(id_)
        return error.code == 0

    def TestComSpeed(self, id_, ch):
        rcvt_speed = c_int32()
        firmware_speed = c_int32()
        self.BL_TestCommSpeed(id_, ch - 1, rcvt_speed, firmware_speed)
        return rcvt_speed.value, firmware_speed.value

    def Disconnect(self, id_):
        self.BL_Disconnect(id_)

    def PluggedChannels(self, id_):
        ch_map = KBIO.ChannelsArray()
        self.BL_GetChannelsPlugged(id_, ch_map, len(ch_map))
        channels = (ch + 1 for ch, present in enumerate(ch_map) if present)
        return channels

    @staticmethod
    def channel_map(channel_set):
        """Build a boolean array of channel presence in the channel_set (an iterable)."""
        channel_map = [False] * max(channel_set)
        for ch in channel_set:
            channel_map[ch - 1] = True
        return channel_map

    def GetChannelInfo(self, id_, ch):
        info = self.ChannelInfo()
        try:
            error = self.BL_GetChannelInfos(id_, ch - 1, info)
        except KBIO_api.BL_Error as e:
            # if firmware is not loaded, part of info is still valid
            # .. so it is not considered an error
            if not e.is_error(KBIO.ERROR.FIRM_FIRMWARENOTLOADED):
                print(f"channel info error : {e}")
        return info

    def LoadFirmware(self, id_, channels, firmware, fpga, force=True):

        results = KBIO.ResultsArray()
        ch_map = KBIO.ChannelsArray(*channels)

        self.BL_LoadFirmware(
            id_,
            ch_map,
            results,
            len(results),
            True,  # display progress bar
            force,
            firmware.encode() if firmware else None,
            fpga.encode() if fpga else None,
        )

        # sift through results and print message in case of error
        for ch, r in enumerate(results):
            error = self.Error(r)
            error.check(f"LoadFirwmare on {ch+1}", abort=False)

    def GetHardwareConf(self, id_, ch):
        conf = self.HardwareConf()
        self.BL_GetHardConf(id_, ch - 1, conf)
        return conf

    def SetHardwareConf(self, id_, ch, cnx, mode):
        hw_conf = self.HardwareConf(cnx, mode)
        self.BL_SetHardConf(id_, ch - 1, hw_conf)

    def OptionError(self, id_, ch):
        code = c_int32()
        pos = c_int32()
        self.BL_GetOptErr(id_, ch - 1, code, pos)
        return code.value, pos.value

    def GetMessage(self, id_, ch):
        message = c_buffer(4096)
        self.BL_GetMessage(id_, ch - 1, *message.parm)
        return message.value

    def GetErrorMsg(self, code):
        message = c_buffer(256)
        error = self.BL_GetErrorMsg(code, *message.parm)
        return message.value

    def DefineParameter(self, label, value, index, parm):
        function = {
            int: self.BL_DefineIntParameter,
            float: self.BL_DefineSglParameter,
            bool: self.BL_DefineBoolParameter,
        }[type(value)]
        function(label.encode(), value, index, parm)

    def DefineBoolParameter(self, label, value, index, parm):
        self.BL_DefineBoolParameter(label.encode(), value, index, parm)

    def DefineSglParameter(self, label, value, index, parm):
        self.BL_DefineSglParameter(label.encode(), value, index, parm)

    def DefineIntParameter(self, label, value, index, parm):
        self.BL_DefineIntParameter(label.encode(), value, index, parm)

    def UpdateParameters(self, id_, ch, index, parms, file):
        self.BL_UpdateParameters(id_, ch - 1, index, parms, file.encode())

    def GetTechniqueInfos(self, id_, ch, ix, info):
        self.BL_GetTechniqueInfos(id_, ch - 1, ix, info)

    def GetParamInfos(self, id_, ch, ix, info):
        self.BL_GetParamInfos(id_, ch - 1, ix, info)

    def LoadTechnique(self, id_, ch, file, parms, first=True, last=True, display=False):
        self.BL_LoadTechnique(id_, ch - 1, file.encode(), parms, first, last, display)

    def StartChannel(self, id_, ch):
        self.BL_StartChannel(id_, ch - 1)

    def StopChannel(self, id_, ch):
        self.BL_StopChannel(id_, ch - 1)

    def StartChannels(self, id_, channels):

        results = KBIO.ResultsArray()
        ch_map = KBIO.ChannelsArray(*channels)
        self.BL_StartChannels(id_, ch_map, results, len(results))

        ok = True
        nb = max(channels)

        # decode the results array and print errors, if any
        for ch, r in enumerate(results):
            if ch >= nb:
                break
            if r != 0:
                ok = False
            error = self.Error(r)
            error.check(f"StartChannels on {ch+1}", abort=False)

        # return whether an error occured
        return ok

    def StopChannels(self, id_, channels):

        results = KBIO.ResultsArray()
        ch_map = KBIO.ChannelsArray(*channels)
        self.BL_StopChannels(id_, ch_map, results, len(results))

        ok = True
        nb = max(channels)

        # decode the results array and print errors, if any
        for ch, r in enumerate(results):
            if ch >= nb:
                break
            if r != 0:
                ok = False
            error = self.Error(r)
            error.check("StopChannels on {ch+1}", abort=False)

        # return whether an error occured
        return ok

    def GetCurrentValues(self, id_, ch):
        cv = KBIO.CurrentValues()
        self.BL_GetCurrentValues(id_, ch - 1, cv)
        return cv

    def GetData(self, id_, ch):

        pb = KBIO.DataBuffer()
        di = KBIO.DataInfo()
        cv = KBIO.CurrentValues()
        self.BL_GetData(id_, ch - 1, pb, di, cv)

        rows = di.NbRows
        cols = di.NbCols
        size = rows * cols
        db = array("L", pb[:size])

        # return CurrentValues, DataInfo, Data Records
        return cv, di, db

    def ConvertNumericIntoSingle(self, vi):
        """Convert the vi word (32b) into a float."""
        vf = c_float()
        self.BL_ConvertNumericIntoSingle(vi, vf)
        return vf.value

    # ==============================================================================#

    def FindEChemDev(self):
        serialized = c_buffer(8192, "UTF16")
        nb_devices = c_uint32()
        self.BL_FindEChemDev(*serialized.parm, nb_devices)
        devices = self._parse_device_serialization(nb_devices.value, serialized.value)
        return devices

    def FindEChemEthDev(self):
        serialized = c_buffer(4096, "UTF16")
        nb_devices = c_uint32()
        self.BL_FindEChemEthDev(*serialized.parm, nb_devices)
        devices = self._parse_device_serialization(nb_devices.value, serialized.value)
        return devices

    def FindEChemUsbDev(self):
        serialized = c_buffer(4096, "UTF16")
        nb_devices = c_uint32()
        self.BL_FindEChemUsbDev(*serialized.parm, nb_devices)
        devices = self._parse_device_serialization(nb_devices.value, serialized.value)
        return devices

    def SetEthernetConfig(self, target_ip, new_ip=None, netmask=None, gateway=None):
        new_config = ""
        if new_ip:
            new_config += f"IP%{new_ip}$"
        if netmask:
            new_config += f"NM%{netmask}$"
        if gateway:
            new_config += f"GW%{gateway}$"
        self.BL_SetConfig(target_ip, new_config)

    # --------------------------------------------------------------------------#

    @classmethod
    def _parse_device_serialization(cls, nb_devices, serialized):

        """Analyze a serialized instrument bundle and turn into a list of devices."""

        devices = list()

        if not serialized:
            return devices

        # check instrument separator is correct

        sep = "%"
        last = serialized[-1]
        separators = (last, sep)

        if not warn_diff("serialization does not end with separator", separators):
            serialized = serialized[:-1]

        instruments = serialized.split(sep)

        for instrument in instruments:

            # separate instrument info into fragments
            all_frags = instrument.split("$")
            mode = all_frags[0]
            # remove blank fields as not meaningful
            fragments = [f for f in all_frags if f]

            if mode == "USB":

                try:
                    # decode fragments into their repective fields
                    index, kind, serial = fragments[1:]
                    index = int(index)
                except:
                    raise RuntimeError(f"ill formed USB serialization ({fragments})")

                # make fields into an USB_device object
                device = KBIO.USB_device(index, kind, serial)

            elif mode == "Ethernet":

                try:
                    # decode fragments into their repective fields :
                    # config = ip_address, gateway, netmask, mac_address
                    *config, identifier, instrument, serial, name = fragments[1:]
                except:
                    raise RuntimeError(
                        f"ill formed Ethernet serialization ({fragments})"
                    )

                # make fields into an Ethernet_device object
                device = KBIO.Ethernet_device(
                    config, instrument, serial, identifier, name
                )

            else:

                raise RuntimeError(f"serialization not understood ({serialized})")

            devices.append(device)

        # check consistency of number of decoded instruments
        nbs = (nb_devices, len(devices))
        warn_diff(f"unexpected nb of devices", nbs)

        return devices

    # --------------------------------------------------------------------------#

    class DeviceInfo(KBIO.DeviceInfo):

        """DeviceInfo adds a few helper methods over the KBIO plain old data equivalent"""

        @property
        def model(self):
            device = KBIO.DEVICE(self.DeviceCode)
            return device.name

        def __str__(self):

            fragments = list()

            channels = self.NumberOfChannels
            slots = self.NumberOfSlots

            fragments.append(
                f"{self.model} {self.RAMSize}MB, CPU={self.CPU}"
                f", {pp_plural(channels,'channel')}"
                f", {pp_plural(slots,'slot')}"
            )
            fragments.append(
                f"Firmware: v{self.FirmwareVersion/100:.2f} "
                f"{self.FirmwareDate_yyyy}/{self.FirmwareDate_mm}/{self.FirmwareDate_dd}"
            )

            cnx = self.NbOfConnectedPC
            fragments.append(
                f"{pp_plural(cnx,'connection')}"
                f", HTdisplay {'on' if self.HTdisplayOn else 'off'}"
            )

            en_clair = "\n".join(fragments)
            return en_clair

    # --------------------------------------------------------------------------#

    class ChannelInfo(KBIO.ChannelInfo):

        """ChannelInfo adds a few helper methods over the KBIO plain old data equivalent"""

        @property
        def firmware(self):
            firmware = KBIO.FIRMWARE(self.FirmwareCode)
            return firmware.name

        @property
        def has_no_firmware(self):
            firmware = KBIO.FIRMWARE(self.FirmwareCode)
            has_no_firmware = firmware.value == 0
            return has_no_firmware

        @property
        def is_kernel_loaded(self):
            firmware = KBIO.FIRMWARE(self.FirmwareCode)
            return firmware.name == "KERNEL"

        @property
        def board(self):
            board = KBIO.CHANNEL_BOARD(self.BoardVersion)
            return board.name

        @property
        def state(self):
            state = KBIO.PROG_STATE(self.State)
            return state.name

        @property
        def amplifier(self):
            amplifier = KBIO.AMPLIFIER(self.AmpCode)
            return amplifier.name

        @property
        def min_IRange(self):
            min_IRange = KBIO.I_RANGE(self.MinIRange)
            return min_IRange.name

        @property
        def max_IRange(self):
            max_IRange = KBIO.I_RANGE(self.MaxIRange)
            return max_IRange.name

        def __str__(self):

            fragments = list()

            if self.has_no_firmware:

                fragments.append(f"{self.board} board, no firmware")

            elif self.is_kernel_loaded:

                fragments.append(f"Channel: {self.Channel+1}")
                fragments.append(f"{self.board} board, S/N {self.BoardSerialNumber}")
                fragments.append(f"{'has a'if self.Lcboard else 'no'} LC head")
                fragments.append(f"{'with' if self.Zboard else 'no'} EIS capabilities")
                fragments.append(pp_plural(self.NbOfTechniques, "technique"))
                fragments.append(f"State: {self.state}")

                if self.NbAmps:
                    fragments.append(f"{self.amplifier} amplifier (x{self.NbAmps})")
                else:
                    fragments.append(f"no amplifiers")

                fragments.append(f"IRange: [{self.min_IRange}, {self.max_IRange}]")
                fragments.append(f"MaxBandwidth: {self.MaxBandwidth}")

                memsize = self.MemSize
                if memsize:
                    fragments.append(
                        f"Memory: {self.MemSize/1024:.1f}KB"
                        f" ({(self.MemFilled/self.MemSize*100.):.2f}% filled)"
                    )
                else:
                    fragments.append("Memory: 0KB")

                version = self.FirmwareVersion / 1000
                vstr = f"{version*10:.2f}" if version < 1.0 else f"{version:.3f}"

                fragments.append(
                    f"{self.firmware} (v{vstr}), " f"FPGA ({self.XilinxVersion:04X})"
                )

            else:

                version = self.FirmwareVersion / 100
                vstr = f"{version*10:.2f}" if version < 1.0 else f"{version:.3f}"
                fragments.append(
                    f"{self.firmware} (v{vstr}), " f"FPGA ({self.XilinxVersion:04X})"
                )

            en_clair = "\n".join(fragments)
            return en_clair

    # --------------------------------------------------------------------------#

    class HardwareConf(KBIO.HardwareConf):

        """HardwareConf adds a few helper methods over the KBIO plain old data equivalent"""

        @property
        def mode(self):
            mode = KBIO.HW_MODE(self.Mode)
            return mode.name

        @property
        def connection(self):
            connection = KBIO.HW_CNX(self.Connection)
            return connection.name

    # --------------------------------------------------------------------------#

    class BL_Error(RuntimeError):

        """BL_Error is an Exception used to capture an EClib API error."""

        def __init__(self, context):
            """Encapsulate context (an Error object)."""
            self.context = context

        def __str__(self):
            return str(self.context)

        def is_error(self, error):
            """Check whether the error code is the same as our error code."""
            is_error = self.context.is_error(error)
            return is_error

    class Error:

        """Class to encapsulate an EClib error code."""

        def __init__(self, code):
            """Encapsulate an error code."""
            self.name = "EClib"
            self.code = code

        @property
        def translate(self):
            """Turn error code into tuple (code, enum-name, clear text)."""
            code = self.code
            try:
                tag = KBIO.ERROR(code)
                en_clair = self.list_by_tag[tag]
                traduction = code, tag.name, en_clair
            except:
                traduction = code, "UNKNOWN_ERROR", "Unknown error"
            return traduction

        def __repr__(self):
            """Full text representation of error"""
            code, tag, description = self.translate
            en_clair = f"{self.name} error {code} [{tag}], {description}"
            return en_clair

        def __str__(self):
            """Clear text representation of error code."""
            code, tag, description = self.translate
            en_clair = f"{description}"
            return en_clair

        def is_error(self, error):
            """Return whether error code is same as numeric value"""
            is_error = self.code == error.value
            return is_error

        def check(self, context=None, abort=True, show=True):
            """Raise an error or print an error in case an error happened.

            context gives local info in case of error, otherwise keep default one
            abort decide between raising an exception versus just printing
            (if show is set to True)
            """
            happened = self.code != 0
            if happened:
                if abort:
                    if context is not None:
                        self.name = context
                    exception = KBIO_api.BL_Error(self)
                    raise exception
                else:
                    if show:
                        print(f"{context} : {self.translate}")

        # ----------------------------------------------------------------------#

        list_by_tag = {
            KBIO.ERROR.NOERROR: "No error",
            KBIO.ERROR.GEN_NOTCONNECTED: "No instrument connected",
            KBIO.ERROR.GEN_CONNECTIONINPROGRESS: "Connection in progress",
            KBIO.ERROR.GEN_CHANNELNOTPLUGGED: "Selected channel(s) unplugged",
            KBIO.ERROR.GEN_INVALIDPARAMETERS: "Invalid function parameters",
            KBIO.ERROR.GEN_FILENOTEXISTS: "Selected file does not exist",
            KBIO.ERROR.GEN_FUNCTIONFAILED: "Function failed",
            KBIO.ERROR.GEN_NOCHANNELSELECTED: "No channel selected",
            KBIO.ERROR.GEN_INVALIDCONF: "Invalid instrument configuration",
            KBIO.ERROR.GEN_ECLAB_LOADED: "EC-Lab firmware loaded on the instrument",
            KBIO.ERROR.GEN_LIBNOTCORRECTLYLOADED: "Library not correctly loaded in memory",
            KBIO.ERROR.GEN_USBLIBRARYERROR: "USB library not correctly loaded in memory",
            KBIO.ERROR.GEN_FUNCTIONINPROGRESS: "Function already in progress",
            KBIO.ERROR.GEN_CHANNEL_RUNNING: "Selected channel(s) already used",
            KBIO.ERROR.GEN_DEVICE_NOTALLOWED: "Device not allowed",
            KBIO.ERROR.GEN_UPDATEPARAMETERS: "Invalid update function parameters",
            KBIO.ERROR.INSTR_VMEERROR: "Internal instrument communication failed",
            KBIO.ERROR.INSTR_TOOMANYDATA: "Too many data to transfer from the instrument (device error)",
            KBIO.ERROR.INSTR_RESPNOTPOSSIBLE: "Selected channel(s) unplugged (device error)",
            KBIO.ERROR.INSTR_RESPERROR: "Instrument response error",
            KBIO.ERROR.INSTR_MSGSIZEERROR: "Invalid message size",
            KBIO.ERROR.COMM_COMMFAILED: "Communication failed with the instrument",
            KBIO.ERROR.COMM_CONNECTIONFAILED: "Cannot establish connection with the instrument",
            KBIO.ERROR.COMM_WAITINGACK: "Waiting for the instrument response",
            KBIO.ERROR.COMM_INVALIDIPADDRESS: "Invalid IP address",
            KBIO.ERROR.COMM_ALLOCMEMFAILED: "Cannot allocate memory in the instrument",
            KBIO.ERROR.COMM_LOADFIRMWAREFAILED: "Cannot load firmware into selected channel(s)",
            KBIO.ERROR.COMM_INCOMPATIBLESERVER: "Communication firmware not compatible",
            KBIO.ERROR.COMM_MAXCONNREACHED: "Maximum number of allowed connections reached",
            KBIO.ERROR.FIRM_FIRMFILENOTEXISTS: "Cannot find kernel.bin file",
            KBIO.ERROR.FIRM_FIRMFILEACCESSFAILED: "Cannot read kernel.bin file",
            KBIO.ERROR.FIRM_FIRMINVALIDFILE: "Invalid kernel.bin file",
            KBIO.ERROR.FIRM_FIRMLOADINGFAILED: "Cannot load kernel.bin on the selected channel(s)",
            KBIO.ERROR.FIRM_XILFILENOTEXISTS: "Cannot find FPGA file",
            KBIO.ERROR.FIRM_XILFILEACCESSFAILED: "Cannot read FPGA file",
            KBIO.ERROR.FIRM_XILINVALIDFILE: "Invalid FPGA file",
            KBIO.ERROR.FIRM_XILLOADINGFAILED: "Cannot load FPGA file on the selected channel(s)",
            KBIO.ERROR.FIRM_FIRMWARENOTLOADED: "No firmware loaded on the selected channel(s)",
            KBIO.ERROR.FIRM_FIRMWAREINCOMPATIBLE: "Loaded firmware not compatible with the library",
            KBIO.ERROR.TECH_ECCFILENOTEXISTS: "Cannot find the selected ECC file",
            KBIO.ERROR.TECH_INCOMPATIBLEECC: "ECC file not compatible with the channel firmware",
            KBIO.ERROR.TECH_ECCFILECORRUPTED: "ECC file corrupted",
            KBIO.ERROR.TECH_LOADTECHNIQUEFAILED: "Cannot load the ECC file",
            KBIO.ERROR.TECH_DATACORRUPTED: "Data returned by the instrument are corrupted",
            KBIO.ERROR.TECH_MEMFULL: "Cannot load techniques: full memory",
            KBIO.ERROR.OPT_CHANGE: "Number of options changed",
            KBIO.ERROR.OPT_4A_ERROR: "4A amplifier unknown error",
            KBIO.ERROR.OPT_4A_OVERTEMP: "4A amplifier temperature overflow",
            KBIO.ERROR.OPT_4A_BADPOWER: "4A amplifier bad power",
            KBIO.ERROR.OPT_4A_POWERFAIL: "4A amplifier power fail",
            KBIO.ERROR.OPT_48V_ERROR: "48V amplifier unknown error",
            KBIO.ERROR.OPT_48V_OVERTEMP: "48V amplifier temperature overflow",
            KBIO.ERROR.OPT_48V_BADPOWER: "48V amplifier bad power",
            KBIO.ERROR.OPT_48V_POWERFAIL: "48V amplifier power fail",
            KBIO.ERROR.OPT_10A5V_ERROR: "10A 5V amplifier error",
            KBIO.ERROR.OPT_10A5V_OVERTEMP: "10A 5V amplifier overheat",
            KBIO.ERROR.OPT_10A5V_BADPOWER: "10A 5V amplifier bad power",
            KBIO.ERROR.OPT_10A5V_POWERFAIL: "10A 5V amplifier power fail",
            KBIO.ERROR.OPT_1A48VP_ERROR: "1A48VP amplifier error",
            KBIO.ERROR.OPT_1A48VP_OVERTEMP: "1A48VP amplifier overheat",
            KBIO.ERROR.OPT_1A48VP_BADPOWER: "1A48VP amplifier bad power",
            KBIO.ERROR.OPT_1A48VP_POWERFAIL: "1A48VP amplifier power fail",
        }

    # ==========================================================================#

    class FindError(BL_Error):

        list_by_tag = {
            KBIO.FIND_ERROR.NO_ERROR: "no error",
            KBIO.FIND_ERROR.UNKNOWN_ERROR: "unknown error",
            KBIO.FIND_ERROR.INVALID_PARAMETER: "invalid function parameters",
            KBIO.FIND_ERROR.ACK_TIMEOUT: "instrument response timeout",
            KBIO.FIND_ERROR.EXP_RUNNING: "experiment is running on instrument",
            KBIO.FIND_ERROR.CMD_FAILED: "instrument do not execute command",
            KBIO.FIND_ERROR.FIND_FAILED: "find failed",
            KBIO.FIND_ERROR.SOCKET_WRITE: "cannot write the request of the descriptions of Ethernet instruments",
            KBIO.FIND_ERROR.SOCKET_READ: "cannot read descriptions of Ethernet instrument",
            KBIO.FIND_ERROR.CFG_MODIFY_FAILED: "set TCP/IP parameters failed",
            KBIO.FIND_ERROR.READ_PARAM_FAILED: "deserialization of TCP/IP parameters failed",
            KBIO.FIND_ERROR.EMPTY_PARAM: "not any TCP/IP parameters in serialization",
            KBIO.FIND_ERROR.IP_FORMAT: "invalid format of IP address",
            KBIO.FIND_ERROR.NM_FORMAT: "invalid format of netmask address",
            KBIO.FIND_ERROR.GW_FORMAT: "invalid format of gateway address",
            KBIO.FIND_ERROR.IP_NOT_FOUND: "instrument to modify not found",
            KBIO.FIND_ERROR.IP_ALREADYEXIST: "new IP address in TCP/IP parameters",
        }

    # ==========================================================================#

    """ List of the EClib entry points, their parameter types and return type if not standard. """

    ecl_api = [
        ("BL_GetLibVersion", [c_char_p, c_uint32_p]),
        ("BL_Connect", [c_char_p, c_uint8, c_int32_p, KBIO.DEVICE_INFO]),
        (
            "BL_GetUSBdeviceinfos",
            [
                c_uint32,
                c_char_p,
                c_uint32_p,
                c_char_p,
                c_uint32_p,
                c_char_p,
                c_uint32_p,
            ],
            c_bool,
        ),
        ("BL_Disconnect", [c_int32]),
        ("BL_TestConnection", [c_int32]),
        ("BL_TestCommSpeed", [c_int32, c_uint8, c_int32_p, c_int32_p]),
        ("BL_GetChannelsPlugged", [c_int32, KBIO.ChannelsArray, c_uint8]),
        (
            "BL_LoadFirmware",
            [
                c_int32,
                KBIO.ChannelsArray,
                KBIO.ResultsArray,
                c_uint8,
                c_bool,
                c_bool,
                c_char_p,
                c_char_p,
            ],
        ),
        ("BL_GetChannelInfos", [c_int32, c_uint8, KBIO.CH_INFO]),
        ("BL_GetHardConf", [c_int32, c_uint8, KBIO.HW_CONF]),
        ("BL_SetHardConf", [c_int32, c_uint8, KBIO.HardwareConf]),
        ("BL_GetErrorMsg", [c_int32, c_char_p, c_uint32_p], int),
        ("BL_GetOptErr", [c_int32, c_int8, c_int32_p, c_int32_p]),
        ("BL_GetMessage", [c_int32, c_uint8, c_char_p, c_uint32_p]),
        (
            "BL_LoadTechnique",
            [c_int32, c_uint8, c_char_p, KBIO.EccParams, c_bool, c_bool, c_bool],
        ),
        ("BL_DefineBoolParameter", [c_char_p, c_bool, c_int32, KBIO.ECC_PARM]),
        ("BL_DefineSglParameter", [c_char_p, c_float, c_int32, KBIO.ECC_PARM]),
        ("BL_DefineIntParameter", [c_char_p, c_int32, c_int32, KBIO.ECC_PARM]),
        ("BL_UpdateParameters", [c_int32, c_int8, c_int32, KBIO.ECC_PARMS, c_char_p]),
        ("BL_GetParamInfos", [c_int32, c_int8, c_int32, KBIO.TECHNIQUE_INFOS]),
        ("BL_GetTechniqueInfos", [c_int32, c_int8, c_int32, KBIO.TECHNIQUE_INFOS]),
        ("BL_StartChannel", [c_int32, c_int8]),
        ("BL_StartChannels", [c_int32, KBIO.ChannelsArray, KBIO.ResultsArray, c_uint8]),
        ("BL_StopChannel", [c_int32, c_int8]),
        ("BL_StopChannels", [c_int32, KBIO.ChannelsArray, KBIO.ResultsArray, c_uint8]),
        ("BL_GetCurrentValues", [c_int32, c_int8, KBIO.CURRENT_VALUES]),
        (
            "BL_GetData",
            [c_int32, c_int8, KBIO.DataBuffer, KBIO.DATA_INFO, KBIO.CURRENT_VALUES],
        ),
        ("BL_ConvertNumericIntoSingle", [c_uint32, c_float_p]),
    ]

    """ List of the blfind entry points, and their parameter types. """

    blfind_api = [
        ("BL_FindEChemDev", [c_char_p, c_uint32_p, c_uint32_p]),
        ("BL_FindEChemEthDev", [c_char_p, c_uint32_p, c_uint32_p]),
        ("BL_FindEChemUsbDev", [c_char_p, c_uint32_p, c_uint32_p]),
        ("BL_SetConfig", [c_char_p, c_char_p]),
    ]

    # --------------------------------------------------------------------------#

    def __init__(self, eclib_file=None, blfind_file=None):
        """Rebuild the api dicts with ctype function with attribute and result types."""

        # first visit the EClib.dll API ..

        try:
            api = self.ecl_api
            dll = WinDLL(eclib_file) if eclib_file else None
        except FileNotFoundError as e:
            raise FileNotFoundError(eclib_file)
        except OSError as e:
            if e.winerror == 193:
                raise RuntimeError(f"{eclib_file} and Python mismatch.")
            else:
                raise

        for name, argtypes, *args in api:
            # rebuild ECLib api dict with ctype func with attributes
            # and optional result type
            self.bind_function(dll, name, argtypes, *args)

        # .. next visit the blfind.dll API

        try:
            api = self.blfind_api
            dll = WinDLL(blfind_file) if blfind_file else None
        except FileNotFoundError as e:
            raise FileNotFoundError(blfind_file)
        except OSError as e:
            if e.winerror == 193:
                raise RuntimeError(f"{blfind_file} and Python mismatch.")
            else:
                raise

        for name, argtypes, *args in api:
            # rebuild blfind api dict with ctype func with attributes
            # and optional result type
            self.bind_function(dll, name, argtypes, *args)

    # --------------------------------------------------------------------------#

    def bind_function(self, dll, name, argtypes, restype=None):
        """Rebind api with wrapped ctype function, registering attribute types and error handling."""

        if dll is None:

            # if missing, force errors on each entry point

            def force_error(*args, abort=True):
                # makeshift error
                error = Error(-999)
                error.check("missing dll", abort)
                return error

            # replace function with force_error
            setattr(self, name, force_error)

        else:

            # retrieve function by name
            function = dll[name]
            # set its argument types
            function.argtypes = argtypes

            # by default function will be error checked on return
            guarded = restype is None
            # set function return type
            function.restype = self.Error if guarded else restype

            if guarded:
                # wrap function with a check of return code ..
                def guarded_call(*args, abort=True):
                    error = function(*args)
                    error.check(name, abort)
                    return error

                # replace function with wrapper function
                setattr(self, name, guarded_call)
            else:
                # .. else set it as the function with argtypes+restype set
                setattr(self, name, function)


# ==============================================================================#



import sqlite3
from typing import Callable
from datetime import datetime, timezone
import os
import logging

log = logging.getLogger(__name__)


def get_db_conn(
    dbpath: str,
    type: str = "sqlite3",
) -> tuple:
    if type == "sqlite3":
        sql = sqlite3
    else:
        raise RuntimeError(f"database type '{type}' unsupported")
    conn = sql.connect(dbpath)
    cur = conn.cursor()
    return conn, cur


def queue_setup(
    dbpath: str,
    type: str = "sqlite3",
) -> None:
    conn, cur = get_db_conn(dbpath, type)
    log.debug(f"attempting to find table 'queue' in '{dbpath}'")
    cur.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='queue';")
    exists = bool(len(cur.fetchall()))
    conn.close()
    if exists:
        log.debug(f"table 'queue' present at '{dbpath}'")
    else:
        log.warning(f"creating a new {type} 'queue' table at '{dbpath}'")
        conn, cur = get_db_conn(dbpath, type)
        cur.execute(
            "CREATE TABLE IF NOT EXISTS queue ("
            "    jobid INTEGER PRIMARY KEY AUTOINCREMENT,"
            "    payload TEXT NOT NULL,"
            "    status TEXT NOT NULL,"
            "    submitted_at TEXT NOT NULL,"
            "    executed_at TEXT,"
            "    completed_at TEXT"
            ");"
        )
        conn.commit()
        conn.close()


def state_setup(
    dbpath: str,
    type: str = "sqlite3",
) -> None:
    conn, cur = get_db_conn(dbpath, type)
    log.debug(f"attempting to find table 'state' in '{dbpath}'")
    cur.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='state';")
    exists = bool(len(cur.fetchall()))
    conn.close()
    if exists:
        log.debug(f"table 'state' present at '{dbpath}'")
    else:
        log.warning(f"creating a new {type} 'state' table at '{dbpath}'")
        conn, cur = get_db_conn(dbpath, type)
        cur.execute(
            "CREATE TABLE IF NOT EXISTS state ("
            "    pipeline TEXT PRIMARY KEY,"
            "    sampleid TEXT,"
            "    ready INTEGER NOT NULL,"
            "    jobid INTEGER,"
            "    pid INTEGER,"
            "    FOREIGN KEY (jobid) REFERENCES queue (jobid)"
            "    );"
        )
        conn.commit()
        conn.close()


def job_set_status(
    dbpath: str,
    st: str,
    jobid: int,
    type: str = "sqlite3",
) -> None:
    conn, cur = get_db_conn(dbpath, type)
    cur.execute(f"UPDATE queue SET status = '{st}' WHERE jobid = {jobid};")
    conn.commit()
    conn.close()


def job_get_info(
    dbpath: str,
    jobid: int,
    type: str = "sqlite3",
) -> tuple:
    conn, cur = get_db_conn(dbpath, type)
    cur.execute(
        "SELECT payload, status, submitted_at, executed_at, completed_at FROM queue "
        f"WHERE jobid = {jobid};"
    )
    ret = cur.fetchone()
    conn.close()
    return ret


def job_set_time(
    dbpath: str,
    tcol: str,
    jobid: int,
    type: str = "sqlite3",
) -> None:
    conn, cur = get_db_conn(dbpath, type)
    ts = str(datetime.now(timezone.utc))
    cur.execute(f"UPDATE queue SET {tcol} = '{ts}' WHERE jobid = {jobid};")
    conn.commit()
    conn.close()


def job_get_all(
    dbpath: str,
    type: str = "sqlite3",
) -> list[tuple]:
    conn, cur = get_db_conn(dbpath, type)
    cur.execute("SELECT jobid, payload, status FROM queue;")
    ret = cur.fetchall()
    conn.close()
    return ret


def pipeline_reset_job(
    dbpath: str,
    pip: str,
    ready: bool = False,
    type: str = "sqlite3",
) -> None:
    conn, cur = get_db_conn(dbpath, type)
    r = int(ready)
    cur.execute(
        f"UPDATE state SET pid = NULL, jobid = NULL, ready = {r} "
        f"WHERE pipeline = '{pip}';"
    )
    conn.commit()
    conn.close()


def pipeline_assign_job(
    dbpath: str,
    pip: str,
    jobid: int,
    pid: int,
    type: str = "sqlite3",
) -> None:
    conn, cur = get_db_conn(dbpath, type)
    cur.execute(
        f"UPDATE state SET pid = {pid}, jobid = {jobid}, ready = 0 "
        f"WHERE pipeline = '{pip}';"
    )
    conn.commit()
    conn.close()


def pipeline_get_running(
    dbpath: str,
    type: str = "sqlite3",
) -> list[tuple]:
    conn, cur = get_db_conn(dbpath, type)
    cur.execute("SELECT pipeline, jobid, pid FROM state WHERE pid IS NOT NULL;")
    ret = cur.fetchall()
    conn.close()
    return ret


def pipeline_get_all(
    dbpath: str,
    type: str = "sqlite3",
) -> list[tuple]:
    conn, cur = get_db_conn(dbpath, type)
    cur.execute("SELECT pipeline FROM state;")
    ret = [i[0] for i in cur.fetchall()]
    conn.close()
    return ret


def pipeline_get_info(
    dbpath: str,
    pip: str,
    type: str = "sqlite3",
) -> tuple:
    conn, cur = get_db_conn(dbpath, type)
    cur.execute(
        f"SELECT sampleid, ready, jobid, pid FROM state WHERE pipeline='{pip}';"
    )
    ret = cur.fetchone()
    conn.close()
    return ret


def pipeline_remove(
    dbpath: str,
    pip: str,
    type: str = "sqlite3",
) -> None:
    conn, cur = get_db_conn(dbpath, type)
    log.warning(f"deleting pipeline '{pip}' from 'state'")
    cur.execute(f"DELETE FROM state WHERE pipeline='{pip}';")
    conn.commit()
    conn.close()


def pipeline_insert(
    dbpath: str,
    pip: str,
    type: str = "sqlite3",
) -> None:
    conn, cur = get_db_conn(dbpath, type)
    log.info(f"creating pipeline '{pip}' in 'state'")
    cur.execute(
        "INSERT INTO state (pipeline, sampleid, jobid, ready)" "VALUES (?, ?, ?, ?);",
        (pip, None, None, 0),
    )
    conn.commit()
    conn.close()


def pipeline_load_sample(
    dbpath: str,
    pip: str,
    sampleid: str,
    type: str = "sqlite3",
) -> None:
    conn, cur = get_db_conn(dbpath, type)
    cur.execute(
        f"UPDATE state SET sampleid = '{sampleid}', ready = 0 "
        f"WHERE pipeline = '{pip}';"
    )
    conn.commit()
    conn.close()


def pipeline_eject_sample(
    dbpath: str,
    pip: str,
    type: str = "sqlite3",
) -> None:
    conn, cur = get_db_conn(dbpath, type)
    cur.execute(
        f"UPDATE state SET sampleid = NULL, ready = 0 " f"WHERE pipeline = '{pip}';"
    )
    conn.commit()
    conn.close()


def queue_payload(
    dbpath: str,
    pstr: str,
    type: str = "sqlite3",
) -> None:
    conn, cur = get_db_conn(dbpath, type)
    log.info(f"inserting a new job into 'state'")
    cur.execute(
        "INSERT INTO queue (payload, status, submitted_at)" "VALUES (?, ?, ?);",
        (pstr, "q", str(datetime.now(timezone.utc))),
    )
    conn.commit()
    conn.close()



import logging
import random

log = logging.getLogger(__name__)
from datetime import datetime, timezone


def get_status(
    address: str,
    channel: int,
    **kwargs: dict,
) -> tuple[float, dict]:
    """
    Get the current status of the device.

    Parameters
    ----------
    address
        Not used

    channel
        Numeric, 1-indexed ID of the channel.

    dllpath
        Path to the BioLogic DLL file.

    Returns
    -------
    timestamp, ready, metadata: tuple[float, bool, dict]
        Returns a tuple containing the timestamp, readiness status, and
        associated metadata.

    """
    dt = datetime.now(timezone.utc)
    return dt.timestamp(), True, {}


def get_data(
    address: str,
    channel: int,
    **kwargs: dict,
) -> tuple[float, dict]:
    """
    Get cached data from the device.

    Parameters
    ----------
    address
        IP address of the potentiostat.

    channel
        Numeric, 1-indexed ID of the channel.

    dllpath
        Path to the BioLogic DLL file.

    Returns
    -------
    timestamp, nrows, data: tuple[float, int, dict]
        Returns a tuple containing the timestamp and associated metadata.

    """
    dt = datetime.now(timezone.utc)
    npoints = random.randint(0, channel) // 2
    points = []
    for i in range(npoints):
        points.append({"value": random.random()*100})
    data = {"data": points}
    return dt.timestamp(), npoints, data


def start_job(
    address: str,
    channel: int,
    **kwargs: dict,
) -> float:
    """
    Start a job on the device.

    Parameters
    ----------
    address
        IP address of the potentiostat.

    channel
        Numeric, 1-indexed ID of the channel.

    dllpath
        Path to the BioLogic DLL file.

    payload
        A protocol describing the techniques to be executed and their order.

    capacity
        The capacity information for the studied battery cell. Only required for
        battery-testing applications or for payloads where currents are specified
        using C or D rates.

    Returns
    -------
    timestamp
        A timestamp corresponding to the start of the job execution.

    """
    dt = datetime.now(timezone.utc)
    return dt.timestamp()


def stop_job(
    address: str,
    channel: int,
    **kwargs: dict,
) -> float:
    """
    Stop a job running on the device.

    This function stops any currently running technique on the specified channel
    of the device. No data is returned.

    Parameters
    ----------
    address
        IP address of the potentiostat.

    channel
        Numeric, 1-indexed ID of the channel.

    dllpath
        Path to the BioLogic DLL file.

    Returns
    -------
    timestamp
        A timestamp corresponding to the start of the job execution.

    """
    dt = datetime.now(timezone.utc)
    return dt.timestamp()



"""
Main module - executables for tomato.

"""
import argparse
import logging
import appdirs
import os
import yaml
import json
import sqlite3
import textwrap
from importlib import metadata
from datetime import datetime, timezone
from typing import Callable

from . import daemon
from . import dbhandler
from . import setlib
from . import ketchup

log = logging.getLogger(__name__)


def _logging_setup(args):
    loglevel = min(max(30 + 10 * (args.quiet - args.verbose), 10), 50)
    logging.basicConfig(level=loglevel)
    log.debug(f"loglevel set to '{logging._levelToName[loglevel]}'")


def _default_parsers() -> tuple[argparse.ArgumentParser]:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--version",
        action="version",
        version=f'%(prog)s version {metadata.version("tomato")}',
    )

    verbose = argparse.ArgumentParser(add_help=False)

    for p in [parser, verbose]:
        p.add_argument(
            "-v",
            "--verbose",
            action="count",
            default=0,
            help="Increase verbosity by one level.",
        )
        p.add_argument(
            "-q",
            "--quiet",
            action="count",
            default=0,
            help="Decrease verbosity by one level.",
        )
    return parser, verbose


def sync_pipelines_to_state(
    pipelines: list,
    dbpath: str,
    type: str = "sqlite3",
) -> None:
    pstate = dbhandler.pipeline_get_all(dbpath, type)
    for pip in pipelines:
        log.debug(f"checking presence of pipeline '{pip['name']}' in 'state'")
        if pip["name"] not in pstate:
            dbhandler.pipeline_insert(dbpath, pip["name"], type)
    pnames = [p["name"] for p in pipelines]
    for pname in pstate:
        if pname not in pnames:
            dbhandler.pipeline_remove(dbpath, pname, type)
    pstate = dbhandler.pipeline_get_all(dbpath, type)


def run_tomato():
    parser, _ = _default_parsers()
    args = parser.parse_args()
    _logging_setup(args)

    dirs = setlib.get_dirs()
    settings = setlib.get_settings(dirs.user_config_dir, dirs.user_data_dir)
    pipelines = setlib.get_pipelines(settings["devices"]["path"])
    log.debug(pipelines)
    log.debug(f"setting up 'queue' table in '{settings['queue']['path']}'")
    dbhandler.queue_setup(settings["queue"]["path"], type=settings["queue"]["type"])
    log.debug(f"setting up 'state' table in '{settings['queue']['path']}'")
    dbhandler.state_setup(settings["state"]["path"], type=settings["state"]["type"])
    sync_pipelines_to_state(
        pipelines, settings["state"]["path"], type=settings["state"]["type"]
    )

    daemon.main_loop(settings, pipelines)


def run_ketchup():
    parser, verbose = _default_parsers()
    subparsers = parser.add_subparsers(dest="subcommand", required=True)

    submit = subparsers.add_parser("submit")
    submit.add_argument(
        "payload",
        help="File containing the payload to be submitted to tomato.",
        default=None,
    )
    submit.set_defaults(func=ketchup.submit)

    status = subparsers.add_parser("status")
    status.add_argument(
        "jobid",
        nargs="?",
        help=(
            "The jobid of the requested job, "
            "or 'queue' for the status of the queue,"
            "or 'state' for the status of pipelines."
        ),
        default="state",
    )
    status.set_defaults(func=ketchup.status)

    stop = subparsers.add_parser("stop")
    stop.add_argument("jobid", help="The jobid of the job to be stopped.", default=None)
    stop.set_defaults(func=ketchup.stop)

    load = subparsers.add_parser("load")
    load.add_argument("sample", help="Name of the sample to be loaded.", default=None)
    load.add_argument(
        "pipeline", help="Name of the pipeline to load the sample to.", default=None
    )
    load.set_defaults(func=ketchup.load)

    eject = subparsers.add_parser("eject")
    eject.add_argument(
        "pipeline", help="Name of the pipeline to eject any sample from.", default=None
    )
    eject.set_defaults(func=ketchup.eject)

    ready = subparsers.add_parser("ready")
    ready.add_argument(
        "pipeline", help="Name of the pipeline to mark as ready.", default=None
    )
    ready.set_defaults(func=ketchup.ready)

    args, extras = parser.parse_known_args()
    args, extras = verbose.parse_known_args(extras, args)
    _logging_setup(args)

    if "func" in args:
        args.func(args)



named_params = {
    "Rest_time_T": float,
    "Duration_step": float,
    "Record_every_dT": float,
    "Record_every_dE": float,
    "Record_every_dI": float,
    "E_Range": int,
    "I_Range": int,
    "Current_step": float,
    "Voltage_step": float,
    "Scan_Rate": float,
    "vs_initial": bool,
    "Test1_Config": int,
    "Test1_Value": float,
    "Test2_Config": int,
    "Test2_Value": float,
    "Test3_Config": int,
    "Test3_Value": float,
    "Exit_Cond": int,
    "N_Cycles": int,
    "Step_number": int,
    "Scan_number": int,
    "loop_N_times": int,
    "protocol_number": int,
    "Begin_measuring_I": float,
    "End_measuring_I": float,
    "Begin_measuring_E": float,
    "End_measuring_E": float,
}


I_ranges = {
    "keep": -1,
    "100 pA": 0,
    "1 nA": 1,
    "10 nA": 2,
    "100 nA": 3,
    "1 uA": 4,
    "10 uA": 5,
    "100 uA": 6,
    "1 mA": 7,
    "10 mA": 8,
    "100 mA": 9,
    "1 A": 10,
    "booster": 11,
    "auto": 12,
}


E_ranges = {
    "+-2.5 V": 0,
    "+-5.0 V": 1,
    "+-10 V": 2,
    "auto": 3,
}

datatypes = {
    "VMP3": {
        "OCV": ["t_high", "t_low", "Ewe", "Ece"],
        "CPLIMIT": ["t_high", "t_low", "Ewe", "I", "cycle"],
        "CALIMIT": ["t_high", "t_low", "Ewe", "I", "cycle"],
        "PDYNLIMIT": ["t_high", "t_low", "Ec", "<I>", "<Ewe>", "cycle"],
        "GDYNLIMIT": ["t_high", "t_low", "Ic", "<I>", "<Ewe>", "cycle"],
    },
    "SP-300": {
        "OCV": ["t_high", "t_low", "Ewe"],
        "CPLIMIT": ["t_high", "t_low", "Ewe", "I", "cycle"],
        "CALIMIT": ["t_high", "t_low", "Ewe", "I", "cycle"],
        "PDYNLIMIT": ["t_high", "t_low", "<I>", "<Ewe>", "cycle"],
        "GDYNLIMIT": ["t_high", "t_low", "<I>", "<Ewe>", "cycle"],
    },
}

techfiles = {
    "VMP3": {
        "open_circuit_voltage": "ocv.ecc",
        "constant_current": "cplimit.ecc",
        "constant_voltage": "calimit.ecc",
        "sweep_voltage": "vscanlimit.ecc",
        "sweep_current": "iscanlimit.ecc",
        "loop": "loop.ecc",
    },
    "SP-300": {
        "open_circuit_voltage": "ocv4.ecc",
        "constant_current": "cplimit4.ecc",
        "constant_voltage": "calimit4.ecc",
        "sweep_voltage": "vscanlimit4.ecc",
        "sweep_current": "iscanlimit4.ecc",
        "loop": "loop4.ecc",
    },
}



""" Bio-Logic OEM package python API.

This module provides helper functions to the ctypes module,
and can be imported as a whole when low level C interfacing is required.
"""

import binascii

from ctypes import (
    c_bool,
    c_byte,
    c_int8,
    c_uint8,
    c_int32,
    c_uint32,
    c_float,
    c_double,
    c_void_p,
    c_char_p,
    Structure,
    POINTER,
    sizeof,
    addressof,
    string_at,
    WinDLL,
)

# ------------------------------------------------------------------------------#

# pointer version of types

c_float_p = POINTER(c_float)
c_int32_p = POINTER(c_int32)
c_uint32_p = POINTER(c_uint32)

# check on the size of pointers
c_is_64b = sizeof(c_void_p) == 8

# ==============================================================================#


def c_dump(cobj):
    """Print both byte version of object and hex dump."""
    l = sizeof(cobj)
    a = addressof(cobj)
    b = string_at(a, l)
    h = binascii.hexlify(b)
    print(b, "\n", h)


# ------------------------------------------------------------------------------#

# helper class to be used when buffer+sizeof are needed


class c_buffer:
    def __init__(self, size, encoding="UTF8"):
        """Initialize a bytes buffer of given size, using a given encoding.

        The buffer goes along its length, which can be adjusted to the actual value during a call.
        """
        self.encoding = encoding
        buffer = self.buffer = bytes(size)
        length = self.length = c_uint32(size)
        self._as_parameter_ = (buffer, length)

    @property
    def parm(self):
        """mimic ctypes style of embedding the parameter, which is a couple."""
        return self._as_parameter_

    @property
    def value(self):
        """Extract the contents of the buffer using the actual length and turn it to a string."""
        buffer, length = self._as_parameter_
        length = length.value
        if self.encoding == "UTF16":
            length *= 2
        value = buffer[:length]
        # make best effort to decode into a string, forgiving errors
        value = value.decode(self.encoding, errors="backslashreplace")
        # remove spurious NULL ending character
        if value[-1:] == b"\0":
            value = value[:-1]
        return value


# ------------------------------------------------------------------------------#


class POD(Structure):

    """ctypes Structure with helper methods."""

    @property
    def keys(self):
        """Reproduce a dict behaviour."""
        keys = (t[0] for t in self._fields_)
        return keys

    def __repr__(self):
        """Return class name and fields one at a line."""
        entries = str(self).split(", ")
        cls = type(self).__name__
        en_clair = f"{cls} :\n  " + "\n  ".join(entries)
        return en_clair

    def __str__(self):
        """Return key-value pairs separated by commas."""
        entries = list()
        for name in self.keys:
            entries.append(f"{name}={getattr(self,name)}")
        en_clair = ", ".join(entries)
        return en_clair

    def __getattr__(self, name):
        """Access Structure fields with nested attributes."""
        i = name.rfind(".")
        if i == -1:
            # Structure should already have provided the first level attribute.
            raise AttributeError(f"{type(self)} has no '{name}' attribute")
        else:
            o = getattr(self, name[:i])
            v = getattr(o, name[i + 1 :])
        return v

    def subset(self, *fields):
        """Create a dict from a selection of Structure fields."""
        subset = dict()
        if len(fields) == 0:
            # no field means all fields
            fields = self.keys
        for name in fields:
            value = getattr(self, name)
            subset += {name: value}
        return subset


# ==============================================================================#
